;
window.W = window.W || {};


(function (undefined) {
    'use strict';

    function animatedScroll (el, distance, duration) {
        var start = Date.now(),
            step = Math.max(distance / (duration / 60), 3),
            tick = function () {
              if (Date.now() - start > duration) {
                // Animation is taking longer than requested
                // scroll the remaining distance at onces
                el.scrollBy(0, distance);
              } else if (distance > 0) {
                el.scrollBy(0, step);
                distance -= step;
                window.requestAnimationFrame(tick)
              }
            }
      
        window.requestAnimationFrame(tick);
    }

    var MyDatetimeFormatter = _.extend({}, Backgrid.CellFormatter.prototype, {
        fromRaw: function (rawValue, model) {
            moment.locale('fr');
            return moment(rawValue).calendar();
        }
    });

    var MyDatetimeCell = Backgrid.DatetimeCell.extend({
        formatter: MyDatetimeFormatter
    });

    var MyUriCell = Backgrid.StringCell.extend({
        render: function () {
            this.$el.empty();
            var rawValue = this.model.get(this.column.get("name"));
            var uri = "/partitions/" + this.model.get("id");
            var formattedValue = this.formatter.fromRaw(rawValue, this.model);
            this.$el.append($("<a>", {
                tabIndex: -1,
                href: uri,
                title: this.title || formattedValue,
                target: this.target
            }).text(formattedValue));
            this.delegateEvents();
            return this;
        }
    });

    var columns = [
        {
            name: "title",
            label: "Partition",
            editable: false, // By default every cell in a column is editable, but *ID* shouldn't be
            cell: MyUriCell
        },
        {
            name: "created_at",
            label: "Créée le",
            cell: MyDatetimeCell
        },
        {
            name: "updated_at",
            label: "Mise à jour le",
            cell: MyDatetimeCell
        },
        {
            name: "tags",
            label: "Tags",
            cell: "string",
        },
        {
            name: "score_type",
            label: "Type",
            cell: "string",
            formatter: _.extend({}, Backgrid.CellFormatter.prototype, {
                fromRaw: function (rawValue, model) {
                    return ["transcription", "prescription", "traduction"][parseInt(rawValue)]
                }
            })
        },
    ];

    /**
     * The queue stores a list of callbacks to be processed at a later time.
     * Entries in the queue can have four states:
     *  unprocessed, processing, done, failed
     * 
     * It is for now only used to handle the file uploads on a save.
     */

    var Queue = function () {
        this.queue = [];
        this.states = { unprocessed: 0, processing: 1, done: 2, failed: 3 };
    }
    
    /**
     * Adds an entry to the queue, the callback should expect two arguments, like
     * a promise: accept and reject. Where accept should be called on succesfull
     * excecution and reject when the action failed.
     * 
     * @param {function} callback
     * @param {*} id (unique) identifier, used on removing entries from the queue
     */
    Queue.prototype.add = function (callback, id) {
        this.queue.push({
            id: id,
            callback: callback,
            state: this.states.unprocessed
        });
    }
    
    /**
     * Returns unprocessed items from the queue
     */
    Queue.prototype.unprocessed = function () {
        return _.where(this.queue, { state: this.states.unprocessed });
    };
    
    /**
     * Starts processing the queue
     * @param {function} callback to be called when all items where processed
     */
    Queue.prototype.process = function (callback) {
        var unprocessed = this.unprocessed(),
            ready       = _.after(unprocessed.length, callback);
        
        _.each(unprocessed, function (entry) {
            entry.state = this.states.processing;
            entry.callback(_.bind(function () {
                    entry.state = this.states.done;
                    ready();
                }, this), _.bind(function () {
                    entry.state = this.states.failed;
                }, this));
        }, this);
    };

    /**
     * Removes an entry with the given id from the queue
     * @param {*} id identifier of the entry
     */    
    Queue.prototype.cancel = function (id) {
        var key = _.findIndex(this.queue, { id: id });
        if (key > -1) {
            delete this.queue[key];
        }
    }

    var queue = new Queue();

    /**
     * Creates attachment model and queues the upload
     * To hace data on the model the file is directly read as a data url,
     * this url is set on the model to be replaced with a proper URL once
     * the model is created/stored on the server.
     * 
     * @param {string} title name of the file / attachment
     * @param {File} file file to upload
     * @param {function} callback called when entry is succesfully excecuted
     */
    function queuedAttachmentUpload(title, file, callback) {
        var model = new W.AttachmentModel({
            title: title,
        });

        var reader = new FileReader();
        reader.addEventListener('load', function () {
            if (!model.get('attachment')) {
                model.set('attachment', reader.result);
            }
        });

        reader.readAsDataURL(file);

        queue.add((function upload (accept, reject) {
            var formData = new FormData();
            // Ensure we have the latest title on submit.
            formData.append('title', model.get('title'));
            formData.append('attachment', file);
        
            $.ajax({
                type: "POST",
                url: '/api/attachments/',
                enctype: 'multipart/form-data',
                data: formData,
                cache: false,
                contentType: false,
                processData: false,
                success: _.bind(function (data) {
                    model.set(data);
                    if (callback) {
                        callback();
                    }
                    accept();
                }, this),
                error: reject,
                dataType: 'json'
            });
        }), model.cid);

        return model;
    }

    /**
     * Removes attachment model when queue is excecuted
     * @param {BackboneModel} model model to remove
     * @param {function} callback [optional] function to excecute when the model is removed
     */
    function queuedAttachmentRemove(model, callback) {
        queue.add((function (accept, reject) {
            var success = function () {
                if (callback) {
                    callback();
                }
                accept();
            }

            if (model.id) {
                model.destroy({
                    success: success,
                    error: reject
                })
            } else {
                uploadQueue.cancel(model.cid);
                success();
            }
        }), model.cid);
    }

    /**
     * Queued creation of the model on the server.
     * @param {BackboneModel} model 
     * @param {function} callback 
     */
    function queuedAttachmentSave(model, callback) {
        queue.add(function (accept, reject) {
            model.save(model.pick('title'), {
                success: function () {
                    if (callback) {
                        callback();
                    }
                    accept();
                },
                error: reject,
                patch: true // necessary to only send the title change on save
            });
        }, model.cid);
    }

    

    var rangeType = {
        undetermined: '0',
        exact: '1',
        minimal: '2',
        range: '3'
    };


    var languages = [
        {
            patt: /en/,
            label: 'English'
        },
        {
            patt: /fr/,
            label: 'Français'
        },
        {
            patt: /nl/,
            label: 'Nederlands'
        },
        {
            patt: /es/,
            label: 'Español'
        },
        {
            patt: /de/,
            label: 'Deutsch'
        }
    ];

    var tags = W.tags = [">", "|", "//", "≥", "||"];

    function getUserLanguage() {
        var l = navigator.language;

        for (var i = 0; i < languages.length; i++) {
            if (l.search(languages[i].patt) > -1) {
                return languages[i].label;
            }
        }
    }

    function formatTimestamp(timestring) {
        var d = new Date(timestring);
        return d.getDate()
            + '-' + (d.getMonth() + 1)
            + '-' + d.getFullYear()
            + ' ' + d.getHours()
            + ':' + d.getMinutes();
    }

    // https://stackoverflow.com/questions/2897619/using-html5-javascript-to-generate-and-save-a-file#4551467
    function download(filename, text) {
        var a = document.createElement('a');
        a.setAttribute('href', 'data:text/plain;charset=utf-8,' + encodeURIComponent(text));
        a.setAttribute('download', filename);

        if (document.createEvent) {
            var event = document.createEvent('MouseEvents');
            event.initEvent('click', true, true);
            a.dispatchEvent(event);
        }
        else {
            a.click();
        }
    }

    function initState() {
        return {
            'collapsed': {},
            'expandLevel': 1
        }
    };

    var state = initState();

    /**
     * Instantiate with at least a label.
     */
    W.RangeField = Backbone.Marionette.View.extend({
        label: '',

        className: 'fieldgroup',

        tagName: 'section',

        ui: {
            'checkbox': '[name="toggle-range-field"]',
            'typeSelect': '[name="type"]',
            'minimal': '[name="minimal"]',
            'exact': '[name="exact"]',
            'rangeMin': '[name="range-min"]',
            'rangeMax': '[name="range-max"]'
        },

        events: {
            'render': 'onRender',
            'change @ui.checkbox': 'onToggle',
            'change @ui.typeSelect': 'onChangeType',
            'change @ui.rangeMin': 'onChangeRangeMin',
            'change @ui.rangeMax': 'onChangeRangeMax',
        },

        template: '#rangefield-template',

        templateContext: function () {
            var val = this.getOption('value');

            return {
                label: this.getOption('label'),
                enabled: (val) ? true : false,
                value: (val) ? val.value : null,
                type: (val) ? val.type : null,
                rangeType: rangeType
            }
        },

        onRender: function () {
            this.onToggle();
            this.onChangeType();
        },

        enabled: function () {
            return this.ui.checkbox.prop('checked');
        },

        onToggle: function (e) {
            if (e) {
                e.stopImmediatePropagation();
            }

            var disabled = !this.enabled();

            this.ui.typeSelect.prop('disabled', disabled);
            this.ui.minimal.prop('disabled', disabled);
            this.ui.exact.prop('disabled', disabled);
            this.ui.rangeMin.prop('disabled', disabled);
            this.ui.rangeMax.prop('disabled', disabled);
        },

        onChangeType: function (e) {
            if (e) {
                e.stopImmediatePropagation();
            }

            var type = this.ui.typeSelect.val();
            this.$('.type-fields').addClass('hidden');
            this.$('[data-range-type="' + type + '"]').removeClass('hidden');
        },

        /**
         * Ensure min is smaller than or equal to max
         */
        onChangeRangeMax: function (e) {
            if (e) {
                e.stopImmediatePropagation();
            }

            var min = parseInt(this.ui.rangeMin.val()), 
                max = parseInt(this.ui.rangeMax.val());

            this.ui.rangeMin.val(Math.min(min, max));
        },

        /**
         * Ensure max bigger than or equal to min
         */
        onChangeRangeMin: function (e) {
            if (e) {
                e.stopImmediatePropagation();
            }

            var min = parseInt(this.ui.rangeMin.val()), 
                max = parseInt(this.ui.rangeMax.val());

            this.ui.rangeMax.val(Math.max(min, max));
        },

        value: function () {
            if (this.enabled()) {
                switch (this.ui.typeSelect.val()) {
                    case rangeType.undetermined:
                        return {
                            type: rangeType.undetermined,
                            value: null
                        };
                    case rangeType.exact:
                        return {
                            type: rangeType.exact,
                            value: parseInt(this.ui.exact.val())
                        };
                    case rangeType.minimal:
                        return {
                            type: rangeType.minimal,
                            value: parseInt(this.ui.minimal.val())
                        };
                    case rangeType.range:
                        var min = parseInt(this.ui.rangeMin.val());
                        var max = parseInt(this.ui.rangeMax.val());

                        return {
                            type: rangeType.range,
                            value: [Math.min(min, max), Math.max(max, min)]
                        };
                }
            }

            return null;
        }
    });


    W.BoucleField = W.RangeField.extend({ template: '#field-boucle-template' });


    W.AlternativeField = W.RangeField.extend({ template: '#field-alternative-template' });

    // Might have a model? A model means saved to the server?
    // Otherwise a more simple object
    // How to differentiate between synced model and file? Marker?
    W.AttachmentField = Backbone.Marionette.View.extend({
        ui: {
            'title': 'input[name="title"]',
            'file': 'input[name="file"]'
        },
        
        events: {
            'click button': 'delete'
        },

        template: '#field-attachment-template',
        
        templateContext () {
            return {
                'hasModel': (this.model)
            }
        },

        delete: function (e) {
            e.preventDefault();
            var r = confirm("Vous êtes sur le point de supprimer ce pièces-jointes. Continuer?");

            if (r) {
                this.model = null;
                this.render();
            }
        },

        value: function () {
            if (this.model) {
                var title = this.ui.title.val();
                return {
                    'new': false,
                    'title': title,
                    'changed': (this.model.get('title') !== title)
                };
            } else {
                return {
                    new: true,
                    title: this.ui.title.val(),
                    file: this.ui.file.get(0).files[0]
                };
            }
        }
    });


    W.Attachment = Backbone.Marionette.View.extend({
        template: '#attachment-template',

        triggers: {
            'click .btn-close': 'attachment:close'
        }
    });

    /**
     * Node editing form
     */
    W.NodeForm = Backbone.Marionette.View.extend({
        tagName: 'form',
        className: 'axis-edit-form',

        template: '#node-form-template',

        ui: {
            'extraFields': '[name="extra-fields"]',
            'close': '[data-name="close"]',
            'submit': '[data-name="submit"]',
        },

        regions: {
            boucle: {
                el: '[data-name="boucle"]',
                replaceElement: true
            },

            alternative: {
                el: '[data-name="alternative"]',
                replaceElement: true
            },

            attachment: {
                el: '[data-name="attachment"]',
                replaceElement: true
            }
        },

        events: {
            'render': 'onRender',
            'click @ui.extraFields legend': 'toggleExtraFields',
            'click @ui.submit': 'submit',
            'change [name="aspect"]': 'toggleAspect',
        },

        triggers: {
            'click @ui.close': 'form:close'
        },

        templateContext: function () {
            return {
                aspectChoices: ["duratif", "itératif", "sémelfactif"],
                alternativeChoices: ["inclusive", "exclusive"],
                tagChoices: tags,
                rangeType: rangeType
            }
        },

        onRender: function () {
            this.toggleAspect();

            var boucleField = new W.BoucleField({
                label: 'boucle',
                value: this.model.get('boucle')
            });

            var alternativeField = new W.AlternativeField({
                label: 'alternative',
                value: this.model.get('alternative')
            });

            var attachmentField = new W.AttachmentField({model: this.model.get('attachment')});

            this.showChildView('boucle', boucleField);
            this.showChildView('alternative', alternativeField);
            this.showChildView('attachment', attachmentField);

            var el = this.el;
            
            // TODO: improve positioning of the form, especially on small screens.

            if (el) {
                window.requestAnimationFrame(function () {
                    var main = document.querySelector('main.main-area').getBoundingClientRect(),
                        form = el.getBoundingClientRect();
          
                    if (form.bottom > main.bottom) {
                        var distance = form.bottom - main.bottom
                        if (form.height > main.height) {
                            distance -= form.height - main.height;
                        }
                        animatedScroll(document.querySelector('main.main-area'), distance, 500);
                    }
                });
            }
        },

        toggleExtraFields: function () {
            var $extraFields = this.ui.extraFields;
            $extraFields.toggleClass('expanded');

            window.requestAnimationFrame(function () {
                var main = document.querySelector('main.main-area'),
                    mainRect = main.getBoundingClientRect(),
                    extraFieldsRect = $extraFields.get(0).getBoundingClientRect();
                
                    if (extraFieldsRect.bottom > mainRect.bottom) {
                    var distance = extraFieldsRect.bottom - mainRect.bottom;
                    animatedScroll(main, distance, 500);
                }
            });
        },

        /**
         * Check whether the aspect checkbox is checked, dis- / enable subfields
         */
        toggleAspect: function (e) {
            if (e) e.stopImmediatePropagation();
            var active = this.isChecked('aspect');
            this.field('aspect-value').prop('disabled', !active);
        },

        /**
         * Return input field with name
         */
        field: function (name) {
            return this.$el.find('[name="' + name + '"]');
        },

        /**
         * Return value for field with name.
         * Integer fields are converted to numbers
         * All other fields are trimmed.
         */
        fieldVal: function (name) {
            var field = this.field(name);

            if (!field) {
                consoles.warn("Can't find field " + name);
            } else {
                if (!field.prop('disabled')) {
                    var value = field.val();
                    if (field.attr('type') == 'number') {
                        return parseInt(value);
                    } else if (value) {
                        return value.trim();
                    } else {
                        return null;
                    }
                }
            }

            return null;
        },

        isChecked: function (name) {
            return this.field(name).prop('checked');
        },

        submit: function (e) {
            e.stopImmediatePropagation();
            e.preventDefault();
            
            // Todo: merge boucle and alternative into one value / datatype:
            // null | [number, number]
            var data = {
                'actant': this.fieldVal('actant'),
                'adresse': this.fieldVal('adresse'),
                'condition': this.fieldVal('condition'),
                'imperative': this.isChecked('imperative'),
                'terme': this.fieldVal('terme'),
                'tag': this.fieldVal('tag'),
                'title': this.fieldVal('title'),
                'module_': this.isChecked('module'),
                'contingent': this.isChecked('contingent'),
                'boucle': this.getChildView('boucle').value(),
                'alternative': this.getChildView('alternative').value(),
                'aspect': (this.isChecked('aspect')) ? this.fieldVal('aspect-value') : null,
                'commandement': this.fieldVal('commandement'),
                'destination': this.fieldVal('destination'),
                'code': this.fieldVal('code'),
                'indications': this.fieldVal('indications'),
                'attachment': this.getChildView('attachment').value()
            };

            this.trigger('form:submit', data);
        }
    });

    W.NodeInlineTitleForm = Backbone.Marionette.View.extend({
        tagName: 'form',
        className: 'axis-inline-title-form',

        template: '#node-inline-form-template',

        ui: {
            'title': '[name="axis-inline-title"]',
            'extraFields': '[name="axis-inline-title"]',
            'submit': '[name="submit"]',
            'close': '[data-name="close"]',
        },

        events: {
            'submit': 'submit',
            'keyup @ui.title': 'keyup',
            'click @ui.submit': 'submit',
            'click @ui.close': 'close',
        },

        onAttach: function () {
            this.ui.title.select();
        },

        keyup: function (e) {
            if (e.keyCode == 27) {
                this.close();
            }
        },

        close: function (e) {
            this.trigger('inlinetitleform:close');
        },

        submit: function (e) {
            e.stopImmediatePropagation();
            e.preventDefault();

            // Todo: merge boucle and alternative into one value / datatype:
            // null | [number, number]
            var data = this.ui.title.val().trim();

            this.trigger('inlinetitleform:submit', data);
        }
    });

    /**
    * Detail page views
    */
    W.TreeNode = Backbone.Marionette.View.extend({
        tagName: 'li',
        className: 'axis',

        attributes: function () {
            return {
                id: "foo-" + this.model.get("title")
            }
        },

        template: '#node-template-vue',

        templateContext: function () {
            return {
                aspectChoices: ["duratif", "itératif", "sémelfactif"],
                alternativeChoices: ["inclusive", "exclusive"],
                tagChoices: tags,
                altSymbol: this.getAltSymbol(),
                hasIndications: this.hasIndications(),
                showTag: this.showTag(),
                rangeType: rangeType,
                hasSublines: (this.model.get('sublines') && this.model.get('sublines').length > 0)
            }
        },

        ui: {
            'title': '.axis-title',
            'titleSpan': '.axis-title-edit'
        },

        regions: {
            tree: {
                el: 'ol',
                replaceElement: true
            },

            form: {
                el: 'form',
                replaceElement: true
            },

            inlineTitleForm: {
                el: '.axis-inline-title-form',
                replaceElement: true
            },

            attachment: {
                el: '.attachment-view',
                replaceElement: true
            }
        },

        events: {
            'click @ui.title': 'toggleClick',
            'click @ui.titleSpan': 'titleClick',
            'click [name="toggle"]': 'toggle',
            'click [name="edit"]': 'toggleForm',
            'click [name="add"]': 'addLine',
            'click [name="delete"]': 'delete',
            'click .btn-attachment': 'showAttachment',
            'change #tag': 'editSiblings',
            'relocate': 'relocate',
            'relocateChild': 'relocateChild'
        },

        childViewEvents: {
            'form:submit': 'update',
            'form:close': 'toggleForm',
            'inlinetitleform:submit': 'updateTitle',
            'inlinetitleform:close': 'closeInlineForm',
            'attachment:close': 'hideAttachment'
        },

        childViewTriggers: {
            'change': 'change'
        },

        initialize: function () {
            this.model.set('boucle', this.getBoucle());
            this.model.unset('boucle_check');
            this.model.unset('boucle_n');
            this.model.unset('boucle_p');
            this.model.set('alternative', this.getAlternative());
            this.model.unset('alternative_n');
            this.model.unset('alternative_p');
            this.model.unset('alternative_mode');
            this.editMode = false;

            if (!(this.model.cid in state.collapsed)) {
                state.collapsed[this.model.cid] = true;
            }

            this.listenTo(this.model, 'change', this.render);
        },

        /**
        * Adapter function between old and new boucle field
        * temporary
        */
        getBoucle: function () {
            var boucle = this.model.get('boucle');
            var boucle_check = this.model.get('boucle_check');
            var boucle_n = this.model.get('boucle_n');
            var boucle_p = this.model.get('boucle_p');

            // console.log('boucle check', this.model.get('boucle_check'), 'n', boucle_n, 'p', boucle_p);

            if (boucle_check) {
                if (boucle_n === boucle_p) {
                    return {
                        type: rangeType.exact,
                        value: parseInt(boucle_n)
                    }
                } else {
                    return {
                        type: rangeType.range,
                        value: [parseInt(boucle_n), parseInt(boucle_p)]
                    }
                }
            } else if (boucle_check !== undefined) {
                return null;
            }

            return boucle;
        },


        /**
        * Adapter function between old and new alternative field
        * temporary
        */
        getAlternative: function () {
            var alternative = this.model.get('alternative');
            var alternative_n = this.model.get('alternative_n');
            var alternative_p = this.model.get('alternative_p');

            if (typeof alternative === 'object') {
                return alternative;
            } else {
                if (alternative_n !== false && alternative_n !== undefined) {
                    if (alternative_n === alternative_p) {
                        return {
                            type: rangeType.exact,
                            value: parseInt(alternative_n)
                        }
                    } else {
                        return {
                            type: rangeType.range,
                            value: [parseInt(alternative_n), parseInt(alternative_p)]
                        }
                    }
                }
            }

            return null;
        },

        /**
         * Only show indications marker if the line has
         * indications, commandement, destination of code
         */
        hasIndications: function () {
            return (
                this.model.get('indications')
                || this.model.get('commandement')
                || this.model.get('destination')
                || this.model.get('code')
            );
        },

        /**
         * Hide tag when it's the root tag or when the parent is
         * in exclusive alternative.
         */
        showTag: function () {
            var parent = this.model.get('parent');

            if (parent) {
                var parent_alternative = parent.get('alternative');
                var sublineCount = parent.get('sublines').models.length;

                if ((parent_alternative
                    && parent_alternative.type === rangeType.exact
                    && parent_alternative.value === 1)
                    || (sublineCount === 1)) {
                    return false;
                }
                else {
                    return true;
                }
            }

            return false;
        },

        toggleForm: function (e) {
            if (e && e.stopImmediatePropagation) {
                e.stopImmediatePropagation();
            }

            var form = this.getChildView('form');

            if (form) {
                this.getRegion('form').reset();
            } else {
                form = new W.NodeForm({ model: this.model });
                this.showChildView('form', form);
            }
        },

        update: function (data) {
            var model = this.model,
                attachmentData = data['attachment'],
                oldAttachment = model.get('attachment'),
                attachmentChanged = false;
            delete data['attachment'];
            var changes = model.changedAttributes(data);

            /*
                There was an attachment on the old model
                    It got removed -> drop old: queuedRemove
                    There is a new one -> drop old: queuedRemove, add new one: queuedUpload
                    The title changed -> update the attachment model, queued save
                    It did not change -> do nothing
                There was no attachement on the old model
                    There is a new one -> upload new one: queuedUpload
                    There is still none -> do nothing
            */
            if (attachmentData.new || attachmentData.changed) {
                if (attachmentData.new) {
                    if (oldAttachment) {
                        attachmentChanged = true;
                        queuedAttachmentRemove(oldAttachment);
                        data['attachment'] = null;
                    }

                    if (attachmentData.file) {
                        attachmentChanged = true;
                        data['attachment'] = queuedAttachmentUpload(attachmentData.title, attachmentData.file);
                    }
                } else if (attachmentData.changed) {
                    attachmentChanged = true;
                    oldAttachment.set('title', attachmentData.title);
                    data['attachment'] = oldAttachment;
                    if (oldAttachment.id) {
                        queuedAttachmentSave(oldAttachment, oldAttachment.cid);
                    }
                }
            }

            if (changes || attachmentChanged) {
                if (data.alternative) {
                    var alt = data.alternative;
                    var minimal = 2;

                    if (alt.type == rangeType.range) {
                        minimal = alt.value[1];
                    } else if (alt.type == rangeType.exact || alt.type == rangeType.minimal) {
                        minimal = alt.value + 1;
                    }

                    var sublines = this.model.get('sublines');

                    while (sublines.length < minimal) {
                        this.model.addSubLine();
                    }

                    this.expand();
                }

                model.set(data);

                if (model.collection && changes && 'tag' in changes) {
                    // Propagate changed tag.
                    model.collection.each(function (m) {
                        if (m.cid !== model.cid) {
                            m.set('tag', data['tag']);
                        }
                    });
                }

                this.trigger('change');
            } else {
                this.toggleForm();
            }
        },

        showInlineForm: function () {
            var form = new W.NodeInlineTitleForm({ model: this.model });
            this.showChildView('inlineTitleForm', form);
            this.ui.titleSpan.hide();
        },

        closeInlineForm: function () {
            this.detachChildView('inlineTitleForm');
            this.ui.titleSpan.show();
        },

        updateTitle: function (title) {
            var changes = this.model.changedAttributes({ 'title': title });

            if (changes) {
                this.model.set('title', title);
                this.trigger('change');
            } else {
                this.closeInlineForm();
            }
        },

        // Triggered by jquery nested sortable when the item is
        relocate: function (event, ui) {
            var parentItem = this.$el.parent().closest("li");
            parentItem.trigger('relocateChild', [this.model, ui.item.index()]);

            return false;
        },

        // Triggered by a subline view beeing relocated
        relocateChild: function (event, model, index) {
            model.get("parent").get("sublines").remove(model);
            this.model.get("sublines").add(model, { at: index });
            this.render();
            this.trigger('change');

            var alts;
            var alternative_mode = model.get("parent").get("alternative_mode");

            if (alternative_mode) {
                var alternative_n = model.get("parent").get('alternative_n');
                var achildViewTriggerslternative_p = model.get("parent").get('alternative_p');

                if (alternative_mode == "exclusive") {
                    alts = "V" + alternative_p;
                } else {
                    alts = "V" + alternative_n + "-" + alternative_p;
                }

                model.set('alternative_symbole', alts);
            }
            return false;
        },

        delete: function (e) {
            // Todo: insert check to see whether there is more than one subline ?
            e.stopImmediatePropagation();
            // Todo: check cross-consistency of axis
            var r = confirm("Vous êtes sur le point de supprimer l’axe " + this.model.get('title') + " et tous ses sous-axes. Continuer?");
            if (r) {
                this.trigger('change');
                this.model.destroy();
            }
        },

        titleClick: function (e) {
            e.stopImmediatePropagation();
            e.preventDefault();

            if (e.altKey || e.ctrlKey || e.metaKey || e.shiftKey) {
                console.log('edit inline', 'alt: ', e.altKey, 'ctrl: ', e.ctrlKey, 'meta: ', e.metaKey, 'shift: ', e.shiftKey);
                this.showInlineForm();
            } else {
                this.toggle();
            }
        },

        toggleClick: function (e) {
            if (e && e.target == this.ui.title.get(0)
                || e.target == this.ui.titleSpan.get(0)) {
                e.stopImmediatePropagation();
                e.preventDefault();
                return this.toggle(e);
            }
        },

        toggle: function (e) {
            if (e) {
                e.stopImmediatePropagation();
                e.preventDefault();
            }

            // inverts the data-collapsed option
            // Uses CSS to toggle visibility

            state.collapsed[this.model.cid] = !state.collapsed[this.model.cid];
            this.$el.attr("data-collapsed", state.collapsed[this.model.cid]);
        },

        collapse: function () {
            state.collapsed[this.model.cid] = true;
            this.$el.attr("data-collapsed", state.collapsed[this.model.cid]);
        },

        expand: function () {
            state.collapsed[this.model.cid] = false;
            this.$el.attr("data-collapsed", state.collapsed[this.model.cid]);
        },

        getAltSymbol: function () {
            var parent = this.model.get("parent");

            if (parent) {
                var alternative = parent.get("alternative");
                var altSymbol = "";

                if (alternative) {
                    if (alternative.type === rangeType.undetermined) {
                        altSymbol = "V<sub>indeterminé</sub>";
                    } else if (alternative.type === rangeType.exact) {
                        if (alternative.value === 1) {
                            altSymbol = "V";
                        } else {
                            altSymbol = "V<sub>" + alternative.value + "</sub>";
                        }
                    } else if (alternative.type === rangeType.minimal) {
                        altSymbol = "V<sub>" + alternative.value + "+</sub>";
                    } else if (alternative.type === rangeType.range) {
                        altSymbol = "V<sub>" + alternative.value[0] + "-" + alternative.value[1] + "</sub>";
                    }
                }

                return altSymbol;
            }

            return "";
        },

        addLine: function (minimal) {
            // AJouter une subline
            // makes sure the line is unfolded to see the new line.
            this.expand();

            var sublines = this.model.get("sublines");

            if (arguments.length < 1 || typeof minimal !== "number") {
                minimal = Math.max(sublines.length + 1, 2);
            }

            while (sublines.length < minimal) {
                this.model.addSubLine();
            }

            this.trigger('change');
            this.render();

            return false;
        },


        toggleAspectEnabled: function (e) {
            e.stopPropagation();
            console.log($('#aspect').get(0).disabled, !(e.target.checked))
            this.$('#aspect').get(0).disabled = !(e.target.checked);
        },

        onExpandUntill: function (level) {
            if (this.model.getLevel() < level) {
                this.expand();
            } else {
                this.collapse();
            }

            var subtree = this.getChildView('tree');
            if (subtree) {
                subtree.triggerMethod('expandUntill', level);
            }
        },

        onRender: function () {
            // attribut des attributs html pour styler en css
            this.$el.attr("data-collapsed", state.collapsed[this.model.cid]);
            this.$el.attr("data-module", this.model.get("module_"));
            this.$el.attr("data-contingent", this.model.get("contingent"));

            var sublines = this.model.get("sublines");

            //show mainline nodes if they are present
            if (sublines.length) {
                var sublinesView = new W.TreeView({ collection: sublines });
                this.$el.attr("data-sublines", "true");
                this.showChildView('tree', sublinesView);
            }
        },
        
        showAttachment: function (e) {
            if (e) {
                e.preventDefault();
                e.stopImmediatePropagation();
            }
            this.showChildView('attachment', (new W.Attachment({model: this.model.get('attachment')} )));
        },

        hideAttachment: function() {
            this.detachChildView('attachment');
        }
    });


    W.TreeView = Backbone.Marionette.CollectionView.extend({
        tagName: 'ol',

        childView: W.TreeNode,

        childViewTriggers: {
            // Propagate change events to the parent
            'change': 'change'
        },

        onExpandUntill: function (level) {
            this.children.each(function (child) {
                child.triggerMethod('expandUntill', level);
            });
        }
    });


    W.EffectifField = W.RangeField.extend({
        template: '#field-effectif-template',
        enabled: function () { return true },
    });


    W.ScoreForm = Backbone.Marionette.View.extend({
        template: '#score-template-form',

        tagName: 'form',

        className: 'score-form',

        ui: {
            'submit': '[data-name="submit"]',
            'close': '[data-name="close"]',
        },

        regions: {
            effectif: {
                el: '[data-name="effectif"]',
                replaceElement: true
            },
        },

        events: {
            'click @ui.submit': 'submit',
            'change [name="score_type"]': 'onChangeScoreType'
        },

        triggers: {
            'click @ui.close': 'form:close'
        },

        templateContext: function () {
            return {
                rangeType: rangeType,
                getUserLanguage: getUserLanguage
            };
        },

        /**
         * Return input field with name
         */
        field: function (name) {
            return this.$el.find('[name="' + name + '"]');
        },
        /**
         * Return value for field with name.
         * Integer fields are converted to numbers
         * All other fields are trimmed.
         */
        fieldVal: function (name) {
            var field = this.field(name);

            if (!field) {
                consoles.warn("Can't find field " + name);
            } else {
                if (!field.prop('disabled')) {
                    var value = field.val();
                    if (field.attr('type') == 'number') {
                        return parseInt(value);
                    } else if (value) {
                        return value.trim();
                    } else {
                        return "";
                    }
                }
            }

            return "";
        },

        onRender: function () {
            var effectifField = new W.EffectifField({
                label: 'effectif',
                value: this.model.get('effectif')
            });

            this.showChildView('effectif', effectifField);
        },

        onChangeScoreType: function (e) {
            if (parseInt(this.fieldVal('score_type')) === 2) {
                this.field('performance_author').prop('disabled', true);
            } else {
                this.field('performance_author').prop('disabled', false);
            }
        },

        submit: function (e) {
            e.stopImmediatePropagation();
            e.preventDefault();

            var data = {
                title: this.fieldVal('title'),
                score_type: parseInt(this.fieldVal('score_type')),
                score_author: this.fieldVal('score_author'),
                performance_author: this.fieldVal('performance_author'),
                presentation: this.fieldVal('presentation'),
                effectif: this.getChildView('effectif').value(),
                stage_set: this.fieldVal('stage_set'),
                duration: this.fieldVal('duration'),
                tags: this.fieldVal('tags').trim().split(/\s*,\s*/),
                language: this.fieldVal('language')
            };

            this.trigger('form:submit', data);
        }

    });


    W.ScoreDetailView = Backbone.Marionette.View.extend({
        // Conditionnal rendering according we are in edit mode or not
        template: '#score-template-vue',

        attributes: {
            id: 'score',
            tabindex: '1'
        },

        regions: {
            mainline: 'ol',

            form: {
                el: '#sidebar form',
                replaceElement: true
            },

            slider: {
                el: '#expand-level-slider',
                replaceElement: true
            },
        },

        ui: {
            'save': '.save',
            'edit': '.edit',
            'expandLevel': '[name="expand-level"]',
            'scoreMeta': '.score-meta',
            'title': '#title',
            'sidebar': '#sidebar'
        },

        events: {
            'click @ui.save': 'save',
            'click @ui.edit': 'toggleForm',
            'click .delete': 'delete',
            'click .duplicate': 'duplicate',
            'click .unfold': 'unfold',
            'click .fold': 'fold',
            'change @ui.expandLevel': 'expandUntill',
            'click [data-name="export"]': 'downloadAsJSON',
            'click @ui.title': 'toggleMeta'
        },

        childViewEvents: {
            'change': 'childChange',
            'level:change': 'expandUntill',
            'form:close': 'toggleForm',
            'form:submit': 'update'
        },

        templateContext: function () {
            return {
                hasChanges: this.hasChanges,
                depth: this.model.getDepth(),
                rangeType: rangeType,
                formatTimestamp: formatTimestamp,
                metaExpanded: this.metaExpanded
            };
        },

        initialize: function () {
            this.hasChanges = false;
            this.metaToggled = false;
            this.mainlineView = null;
            this.sliderView = null;
            // this.listenTo(this.model, 'change', this.render);
            // this.listenTo(this.model, 'sync', this.render);

            $(window).on('beforeunload', (function (e) {
                if (this.hasChanges) {
                    if (!window.confirm("There are unsaved changes to the score. Are you sure to leave?")) {
                        e.preventDefault();
                        return false;
                    }
                }
            }).bind(this));

            this.listenToOnce(this.model, 'sync', (function () {
                this.listenTo(this.model, 'change', this.render);
                // this.listenTo(this.model, 'sync', this.render);
            }).bind(this));
        },

        toggleForm: function (e) {
            if (e && e.stopImmediatePropagation) {
                e.stopImmediatePropagation();
            }

            var form = this.getChildView('form');

            if (form) {
                this.getRegion('form').reset();
            } else {
                form = new W.ScoreForm({ model: this.model });
                this.showChildView('form', form);
            }
        },

        update: function (data) {
            if (this.model.changedAttributes(data)) {
                this.childChange();
                this.model.set(data);
            } else {
                this.toggleForm();
            }
        },

        childChange: function () {
            this.ui.save.prop('disabled', false);
            this.hasChanges = true;
        },

        unfold: function (view, event) {
            $('li').attr('data-collapsed', false);
            this.getChildView('slider').val(this.model.getDepth());
        },

        fold: function (view, event) {
            $('li').attr('data-collapsed', true);
            this.getChildView('slider').val(1);
        },

        save: function (view, event) {
            var button = this.ui.save;
            this.hasChanges = false;
            button.prop('disabled', true);

            var done = _.bind(function () {
                button.text(button.data('label'));
            }, this);

            var failed = _.bind(function (model, status) {
                button.text(button.data('label'));
                button.prop('disabled', false);
                this.hasChanges = true;
                alert('Error while saving -- ' + status.status + ' -- ' + status.statusText);
            }, this);

            var saveModel = _.bind(function () {
                button.text(button.data('label-progress'));
                this.model.save(null, {
                    silent: true,
                    success: done,
                    error: failed
                });
            }, this);

            if (queue.unprocessed().length > 0) {
                button.text(button.data('label-attachment-process'));
                queue.process(saveModel);
            } else {
                saveModel();
            }
        },

        delete: function (view, event) {
            var r = confirm("Vous êtes sur le point de supprimer la partition intitulée \"" + this.model.get("title") + "\". Continuer?");
            if (r) {
                this.model.destroy({
                    success: function () {
                        Backbone.history.navigate('/', {
                            trigger: true
                        });
                    }
                });
            }
        },

        duplicate: function (view, event) {
            var data = this.model.attributes;
            delete data.id

            var newModel = new W.ScoreModel(data);

            newModel.save({
                title: newModel.get("title") + " (copy)",
            }, {
                    success: function () {
                        Backbone.history.navigate('/partitions/' + newModel.id, {
                            trigger: true
                        });
                    }
                });
        },

        expandUntill: function (level) {
            this.getChildView('mainline').triggerMethod('expandUntill', level);
        },

        onRender: function () {
            document.title = this.model.get('title');

            var mainlineView = new W.TreeNode({ model: this.model.get('mainline') });
            this.showChildView('mainline', mainlineView);

            var sliderView = new W.SliderView({ depth: this.model.getDepth() });
            this.showChildView('slider', sliderView);

            this.getRegion('mainline').$el.nestedSortable({
                placeholder: 'placeholder',
                forcePlaceholderSize: true,
                helper: 'clone',
                handle: '.icon--handle',
                isTree: true,
                items: 'li',
                toleranceElement: '> div',
                protectRoot: true,
                rootID: 'foo',
                isAllowed: function (placeholder, placeholderParent, currentItem) {
                    return true;
                },
                relocate: function (event, ui) {
                    /*
                     * When the sort action is over:
                     * 1. trigger an event "relocate" on the item beeing sorted, along with its index
                     * 2. listen and catch event "relocate" on the view that is responsible for the item
                     * 3. trigger an event "relocateChild" on the parent element, along with the index, and the item model
                     * 4. listen and catch event "relocateChild" on the view that is responsible for the parent item
                     * 5. console.log that
                     */
                    ui.item.trigger("relocate", ui);
                },
            });
        },

        onBeforeDestroy: function () {
            $(window).off('beforeunload');
        },

        downloadAsJSON: function () {
            download(this.model.get('title') + '-export.json', JSON.stringify(this.model.toJSON()));
        },

        toggleMeta: function () {
            this.metaExpanded = !this.metaExpanded;
            this.ui.sidebar.attr('data-expanded', this.metaExpanded);
        }
    });


    W.SliderView = Backbone.Marionette.View.extend({
        template: '#slider-template',

        tagName: 'div',

        className: 'expand-level',

        triggers: {
            'click .expand-level-option': 'click',
        },

        templateContext: function () {
            return {
                'depth': this.getOption('depth')
            };
        },

        onClick: function (view, e) {
            e.preventDefault();
            e.stopImmediatePropagation();
            this.val(parseInt($(e.target).attr('data-value')));
        },

        val: function (level) {
            state.expandLevel = level;
            this.$('[data-selected="true"]').attr('data-selected', false);
            this.$('[data-value="' + level + '"]').attr('data-selected', true);
            this.trigger('level:change', level);
        },

        onRender: function () {
            this.$('[data-selected="true"]').attr('data-selected', false);
            this.$('[data-value="' + state.expandLevel + '"]').attr('data-selected', true);
        }
    });


    /**
     * List page views
     */


    W.SearchView = Backbone.Marionette.View.extend({
        template: '#search-template',

        triggers: {
            'click .search': 'search',
        },

        onSearch: function (view, event) {
            this.collection.queryParams.search = this.$el.find('.search-input').val();
            this.collection.fetch();
        },
    });


    W.FilterView = Backbone.Marionette.View.extend({
        template: '#filters-template',

        triggers: {
            'click a': 'filter',
        },

        onFilter: function (view, event) {
            console.log(event);
            this.collection.queryParams.score_type = $(event.currentTarget).data('value');
            // this.collection.queryParams.score_type = e.val();
            this.collection.fetch();
        },
    });


    W.CreateView = Backbone.Marionette.View.extend({
        template: '#create-template',


        triggers: {
            'submit .create-form': 'create'
        },


        onCreate: function (event) {
            var title = this.$el.find('[name="title"]').val();
            var myModel = new W.ScoreModel();
            myModel.save({
                title: title
            }, {
                    success: function () {
                        Backbone.history.navigate('/partitions/' + myModel.id, {
                            trigger: true
                        });
                    }
                });
        },
    });


    W.ScoreListView = Backbone.Marionette.View.extend({
        template: '#score-list-template',

        attributes: { class: 'score-list' },

        regions: {
            search: '#search',
            list: '#list',
            pagination: '#pagination',
            filters: '#filters',
        },

        onRender: function () {
            document.title = 'Notation W';

            var scoreCollection = new W.ScoreCollection([], {
                // All the `state` and `queryParams` key value pairs are merged with
                // the defaults too.
                state: {
                    pageSize: 25,
                    firstPage: 1,
                    currentPage: 1,
                    sortKey: "title",
                    order: -1
                }
            });

            var myListView = new Backgrid.Grid({
                columns: columns,
                collection: scoreCollection
            });

            var myFilterView = new W.FilterView({
                collection: scoreCollection
            });

            var mySearchView = new W.SearchView({
                collection: scoreCollection
            });

            var myPaginationView = new Backgrid.Extension.Paginator({
                controls : {
                    rewind: {
                        label: "《",
                        title: "First"
                    },
                    back: {
                        label: "〈",
                        title: "Previous"
                    },
                    forward: {
                        label: "〉",
                        title: "Next"
                    },
                    fastForward: {
                        label: "》",
                        title: "Last"
                    }
                },

                // If you anticipate a large number of pages, you can adjust
                // the number of page handles to show. The sliding window
                // will automatically show the next set of page handles when
                // you click next at the end of a window.
                windowSize: 20, // Default is 10

                // Used to multiple windowSize to yield a number of pages to slide,
                // in the case the number is 5
                slideScale: 0.25, // Default is 0.5

                // Whether sorting should go back to the first page
                goBackFirstOnSort: false, // Default is true

                collection: scoreCollection
            });

            var that = this;

            myListView.collection.fetch({
                success: function () {
                    that.showChildView('search', mySearchView);
                    that.showChildView('filters', myFilterView);
                    that.showChildView('list', myListView);
                    that.showChildView('pagination', myPaginationView);
                }
            });
        }
    });


    /**
     * User views
     */


    W.UserListView = Backbone.Marionette.View.extend({
        template: '#user-list-template',
    });


    W.UserDetailView = Backbone.Marionette.View.extend({
        template: '#user-detail-template',
    });


    W.LoginView = Backbone.Marionette.View.extend({
        template: '#login-template',

        ui: {
            'close': '.btn-close',
            'submit': '.btn-submit'
        },

        triggers: {
            'click @ui.close': 'hide:modal',
            'click @ui.submit': 'submit',
        },

        onSubmit: function (foo, event) {
            event.preventDefault();

            var model = this.model;
            var that = this;

            var form = this.$el.find("form");
            var data = form.serialize();

            $.ajax({
                type: "POST",
                url: form.attr('action'),
                data: form.serialize(),
                success: function(data, textStatus, jqXHR) {
                    model.fetch();
                    that.trigger('hide:modal');
                },
                error: function(xhr, ajaxOptions, thrownError){
                    alert('login failed - please try again');
                },
            });
        }
    });


    W.LogoutView = Backbone.Marionette.View.extend({
        template: '#logout-template',

        ui: {
            'close': '.btn-close',
            'submit': '.btn-submit'
        },

        triggers: {
            'click @ui.close': 'hide:modal',
            'click @ui.submit': 'submit'
        },

        onSubmit: function (foo, event) {
            event.preventDefault();
            var that = this;

            var model = this.model;
            var form = this.$el.find("form");

            $.ajax({
                type: "POST",
                url: form.attr('action'),
                success: function(data, textStatus, jqXHR){
                    that.trigger('hide:modal');
                    alert("ok");
                    model.set(model.defaults);
                },
                error: function(xhr, ajaxOptions, thrownError){
                    alert('logout failed - please try again');
                    // that.trigger('hide:modal');
                    // model.clear();
                },
            });
        }
    });


    W.RegisterView = Backbone.Marionette.View.extend({
        template: '#register-template',

        ui: {
            'close': '.btn-close'
        },

        triggers: {
            'click @ui.close': 'hide:modal',
        },
    });


    /**
     *  Other views
     */


    W.AboutView = Backbone.Marionette.View.extend({
        template: '#about-template',

        attributes: { class: 'body-text body-text--about' }
    });


    W.HomeView = Backbone.Marionette.View.extend({
        template: '#home-template',

        regions: {
            latest_updated: '#latest_updated'
        },

        onRender: function () {
            var that = this;

            document.title = 'Notation W';

            var scoreCollection = new W.ScoreCollection([], {
                url: '/test',
                state: {
                    // pageSize: 10,
                    // firstPage: 1,
                    // currentPage: 1,
                    sortKey: "updated_at",
                    order: 1,
                }
            });

            scoreCollection.queryParams.is_featured=true;

            var myListView = new Backgrid.Grid({
                columns: columns,
                collection: scoreCollection
            });

            myListView.collection.fetch({
                success: function () {
                    that.showChildView('latest_updated', myListView);
                }
            });
        }
    });


    W.ToggableView = Backbone.Marionette.View.extend({
        onShow: function () {
            this.$el.removeClass('is-collapsed');
        },

        onHide: function () {
            console.log("hiding");
            this.$el.addClass('is-collapsed');
        },

        onToggle: function () {
            this.$el.toggleClass('is-collapsed');
        },
    });


    W.helpView = W.ToggableView.extend({
        template: '#help-template',

        attributes: {
            id: 'help',
            class: 'is-collapsed'
        },

        ui: {
            'toggle': '.btn-hamburger',
        },

        triggers: {
            'click @ui.show': 'show',
            'click @ui.hide': 'hide',
            'click @ui.toggle': 'toggle',
        },
    });


    W.HeaderView = W.ToggableView.extend({
        tagName: 'header',
        template: '#header-template',

        attributes: {
            class: 'main-header'
        },

        ui: {
            'toggle': '.btn-hamburger',
            'login': '.btn-login',
            'logout': '.btn-logout',
            'register': '.btn-register',
        },

        triggers: {
            'click @ui.show': 'show',
            'click @ui.hide': 'hide',
            'click @ui.toggle': 'toggle',
            'click @ui.login': 'show:login',
            'click @ui.logout': 'show:logout',
            'click @ui.register': 'show:register',
        },

        initialize: function () {
            this.listenTo(this.model, 'change', this.render);
        },

        regions: {
            create: '#create'
        },

        onRender: function () {
            var myCreateView = new W.CreateView();

            this.showChildView('create', myCreateView);
        }
    });


    /**
     *  Base view, the mother of all views
     */

    W.BaseView = Backbone.Marionette.View.extend({
        template: '#base-template',

        attributes: {
            class: 'outer-wrapper',
        },

        regions: {
            header: {
                el: '.main-header',
                replaceElement: true
            },
            main: {
                el: '.main-area',
                // replaceElement: true
            },
            help: {
                el: '#help',
                replaceElement: true
            },
            modal: {
                el: '.modal',
                // replaceElement: true
            }
        },

		childViewEvents: {
			'show:login': 'showLogin',
			'show:register': 'showRegister',
			'show:logout': 'showLogout',
			'hide:modal': 'hideModal'
		},

        onRender: function () {
            this.showChildView('header', new W.HeaderView({model: this.model}));
            this.showChildView('help', new W.helpView());
        },

        showLogin: function(event) {
            this.showChildView('modal', new W.LoginView({model: this.model}));
        },

        showLogout: function(event) {
            this.showChildView('modal', new W.LogoutView({model: this.model}));
        },

        showRegister: function(event) {
            this.showChildView('modal', new W.RegisterView());
        },

        hideModal: function(event) {
            this.detachChildView('modal');
        },
    });
})();
